@@ -5,38 +5,41 @@
 
 import argparse
 import copy
+import json
 import os
 import re
 import sys
-import json
+import typing as t
 import warnings
 
+from traitlets.traitlets import Any, Container, Dict, HasTraits, List, Undefined
+
 from ..utils import cast_unicode, filefind
 
-from traitlets.traitlets import (
-    HasTraits, Container, List, Dict, Any, Undefined,
-)
-
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 # Exceptions
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 
 
 class ConfigError(Exception):
     pass
 
+
 class ConfigLoaderError(ConfigError):
     pass
 
+
 class ConfigFileNotFound(ConfigError):
     pass
 
+
 class ArgumentError(ConfigLoaderError):
     pass
 
-#-----------------------------------------------------------------------------
+
+# -----------------------------------------------------------------------------
 # Argparse fix
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 
 # Unfortunately argparse by default prints help messages to stderr instead of
 # stdout.  This makes it annoying to capture long help screens at the command
@@ -62,17 +65,20 @@ class ArgumentParser(argparse.ArgumentParser):
     def print_help(self, file=None):
         if file is None:
             file = sys.stdout
-        return super(ArgumentParser, self).print_help(file)
+        return super().print_help(file)
 
     print_help.__doc__ = argparse.ArgumentParser.print_help.__doc__
 
-#-----------------------------------------------------------------------------
+
+# -----------------------------------------------------------------------------
 # Config class for holding config information
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
+
 
 def execfile(fname, glob):
-    with open(fname, 'rb') as f:
-        exec(compile(f.read(), fname, 'exec'), glob, glob)
+    with open(fname, "rb") as f:
+        exec(compile(f.read(), fname, "exec"), glob, glob)
+
 
 class LazyConfigValue(HasTraits):
     """Proxy object for exposing methods on configurable containers
@@ -106,7 +112,6 @@ class LazyConfigValue(HasTraits):
         """like list.extend, but for the front"""
         self._prepend[:0] = other
 
-
     def merge_into(self, other):
         """
         Merge with another earlier LazyConfigValue or an earlier container.
@@ -193,31 +198,31 @@ class LazyConfigValue(HasTraits):
         """
         d = {}
         if self._update:
-            d['update'] = self._update
+            d["update"] = self._update
         if self._extend:
-            d['extend'] = self._extend
+            d["extend"] = self._extend
         if self._prepend:
-            d['prepend'] = self._prepend
+            d["prepend"] = self._prepend
         elif self._inserts:
-            d['inserts'] = self._inserts
+            d["inserts"] = self._inserts
         return d
 
     def __repr__(self):
         if self._value is not None:
-            return "<%s value=%r>" % (self.__class__.__name__, self._value)
+            return f"<{self.__class__.__name__} value={self._value!r}>"
         else:
-            return "<%s %r>" % (self.__class__.__name__, self.to_dict())
+            return f"<{self.__class__.__name__} {self.to_dict()!r}>"
 
 
 def _is_section_key(key):
     """Is a Config key a section name (does it start with a capital)?"""
-    if key and key[0].upper()==key[0] and not key.startswith('_'):
+    if key and key[0].upper() == key[0] and not key.startswith("_"):
         return True
     else:
         return False
 
 
-class Config(dict):
+class Config(dict):  # type:ignore[type-arg]
     """An attribute-based dict that can do smart merges.
 
     Accessing a field on a config object for the first time populates the key
@@ -243,9 +248,7 @@ class Config(dict):
         """
         for key in self:
             obj = self[key]
-            if _is_section_key(key) \
-                    and isinstance(obj, dict) \
-                    and not isinstance(obj, Config):
+            if _is_section_key(key) and isinstance(obj, dict) and not isinstance(obj, Config):
                 setattr(self, key, Config(obj))
 
     def _merge(self, other):
@@ -258,7 +261,7 @@ class Config(dict):
         for k, v in other.items():
             if k not in self:
                 to_update[k] = v
-            else: # I have this key
+            else:  # I have this key
                 if isinstance(v, Config) and isinstance(self[k], Config):
                     # Recursively merge common sub Configs
                     self[k].merge(v)
@@ -270,7 +273,7 @@ class Config(dict):
 
         self.update(to_update)
 
-    def collisions(self, other):
+    def collisions(self, other: "Config") -> t.Dict[str, t.Any]:
         """Check for collisions between two config objects.
 
         Returns a dict of the form {"Class": {"trait": "collision message"}}`,
@@ -278,7 +281,7 @@ class Config(dict):
 
         An empty dict indicates no collisions.
         """
-        collisions = {}
+        collisions: t.Dict[str, t.Any] = {}
         for section in self:
             if section not in other:
                 continue
@@ -287,18 +290,18 @@ class Config(dict):
             for key in mine:
                 if key in theirs and mine[key] != theirs[key]:
                     collisions.setdefault(section, {})
-                    collisions[section][key] = "%r ignored, using %r" % (mine[key], theirs[key])
+                    collisions[section][key] = f"{mine[key]!r} ignored, using {theirs[key]!r}"
         return collisions
 
     def __contains__(self, key):
         # allow nested contains of the form `"Section.key" in config`
-        if '.' in key:
-            first, remainder = key.split('.', 1)
+        if "." in key:
+            first, remainder = key.split(".", 1)
             if first not in self:
                 return False
             return remainder in self[first]
 
-        return super(Config, self).__contains__(key)
+        return super().__contains__(key)
 
     # .has_key is deprecated for dictionaries.
     has_key = __contains__
@@ -332,7 +335,7 @@ class Config(dict):
                 c = Config()
                 dict.__setitem__(self, key, c)
                 return c
-            elif not key.startswith('_'):
+            elif not key.startswith("_"):
                 # undefined, create lazy value, used for container methods
                 v = LazyConfigValue()
                 dict.__setitem__(self, key, v)
@@ -343,20 +346,22 @@ class Config(dict):
     def __setitem__(self, key, value):
         if _is_section_key(key):
             if not isinstance(value, Config):
-                raise ValueError('values whose keys begin with an uppercase '
-                                 'char must be Config instances: %r, %r' % (key, value))
+                raise ValueError(
+                    "values whose keys begin with an uppercase "
+                    "char must be Config instances: %r, %r" % (key, value)
+                )
         dict.__setitem__(self, key, value)
 
     def __getattr__(self, key):
-        if key.startswith('__'):
-            return dict.__getattr__(self, key)
+        if key.startswith("__"):
+            return dict.__getattr__(self, key)  # type:ignore[attr-defined]
         try:
             return self.__getitem__(key)
         except KeyError as e:
             raise AttributeError(e)
 
     def __setattr__(self, key, value):
-        if key.startswith('__'):
+        if key.startswith("__"):
             return dict.__setattr__(self, key, value)
         try:
             self.__setitem__(key, value)
@@ -364,7 +369,7 @@ class Config(dict):
             raise AttributeError(e)
 
     def __delattr__(self, key):
-        if key.startswith('__'):
+        if key.startswith("__"):
             return dict.__delattr__(self, key)
         try:
             dict.__delitem__(self, key)
@@ -374,6 +379,7 @@ class Config(dict):
 
 class DeferredConfig:
     """Class for deferred-evaluation of config from CLI"""
+
     pass
 
     def get_value(self, trait):
@@ -401,6 +407,7 @@ class DeferredConfigString(str, DeferredConfig):
 
     .. versionadded:: 5.0
     """
+
     def get_value(self, trait):
         """Get the value stored in this string"""
         s = str(self)
@@ -413,10 +420,10 @@ class DeferredConfigString(str, DeferredConfig):
             return s
 
     def __repr__(self):
-        return '%s(%s)' % (self.__class__.__name__, self._super_repr())
+        return f"{self.__class__.__name__}({self._super_repr()})"
 
 
-class DeferredConfigList(list, DeferredConfig):
+class DeferredConfigList(list, DeferredConfig):  # type:ignore[type-arg]
     """Config value for loading config from a list of strings
 
     Interpretation is deferred until it is loaded into the trait.
@@ -431,6 +438,7 @@ class DeferredConfigList(list, DeferredConfig):
 
     .. versionadded:: 5.0
     """
+
     def get_value(self, trait):
         """Get the value stored in this string"""
         if hasattr(trait, "from_string_list"):
@@ -439,7 +447,9 @@ class DeferredConfigList(list, DeferredConfig):
         else:
             # only allow one item
             if len(self) > 1:
-                raise ValueError(f"{trait.name} only accepts one value, got {len(self)}: {list(self)}")
+                raise ValueError(
+                    f"{trait.name} only accepts one value, got {len(self)}: {list(self)}"
+                )
             src = self[0]
             cast = trait.from_string
 
@@ -452,15 +462,15 @@ class DeferredConfigList(list, DeferredConfig):
             return src
 
     def __repr__(self):
-        return '%s(%s)' % (self.__class__.__name__, self._super_repr())
+        return f"{self.__class__.__name__}({self._super_repr()})"
 
 
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 # Config loading classes
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 
 
-class ConfigLoader(object):
+class ConfigLoader:
     """A object for loading configurations from just about anywhere.
 
     The resulting configuration is packaged as a :class:`Config`.
@@ -477,6 +487,7 @@ class ConfigLoader(object):
 
     def _log_default(self):
         from traitlets.log import get_logger
+
         return get_logger()
 
     def __init__(self, log=None):
@@ -496,7 +507,7 @@ class ConfigLoader(object):
         self.clear()
         if log is None:
             self.log = self._log_default()
-            self.log.debug('Using default logger')
+            self.log.debug("Using default logger")
         else:
             self.log = log
 
@@ -532,15 +543,16 @@ class FileConfigLoader(ConfigLoader):
             The path to search for the config file on, or a sequence of
             paths to try in order.
         """
-        super(FileConfigLoader, self).__init__(**kw)
+        super().__init__(**kw)
         self.filename = filename
         self.path = path
-        self.full_filename = ''
+        self.full_filename = ""
 
     def _find_file(self):
         """Try to find the file by searching the paths."""
         self.full_filename = filefind(self.filename, self.path)
 
+
 class JSONFileConfigLoader(FileConfigLoader):
     """A JSON file loader for config
 
@@ -558,7 +570,7 @@ class JSONFileConfigLoader(FileConfigLoader):
         self.clear()
         try:
             self._find_file()
-        except IOError as e:
+        except OSError as e:
             raise ConfigFileNotFound(str(e))
         dct = self._read_file_as_dict()
         self.config = self._convert_to_config(dct)
@@ -569,15 +581,15 @@ class JSONFileConfigLoader(FileConfigLoader):
             return json.load(f)
 
     def _convert_to_config(self, dictionary):
-        if 'version' in dictionary:
-            version = dictionary.pop('version')
+        if "version" in dictionary:
+            version = dictionary.pop("version")
         else:
             version = 1
 
         if version == 1:
             return Config(dictionary)
         else:
-            raise ValueError('Unknown version of JSON config file: {version}'.format(version=version))
+            raise ValueError(f"Unknown version of JSON config file: {version}")
 
     def __enter__(self):
         self.load_config()
@@ -592,11 +604,10 @@ class JSONFileConfigLoader(FileConfigLoader):
         """
         self.config.version = 1
         json_config = json.dumps(self.config, indent=2)
-        with open(self.full_filename, 'w') as f:
+        with open(self.full_filename, "w") as f:
             f.write(json_config)
 
 
-
 class PyFileConfigLoader(FileConfigLoader):
     """A config loader for pure python files.
 
@@ -609,7 +620,7 @@ class PyFileConfigLoader(FileConfigLoader):
         self.clear()
         try:
             self._find_file()
-        except IOError as e:
+        except OSError as e:
             raise ConfigFileNotFound(str(e))
         self._read_file_as_dict()
         return self.config
@@ -631,6 +642,7 @@ class PyFileConfigLoader(FileConfigLoader):
 
     def _read_file_as_dict(self):
         """Load the config file into self.config, with recursive loading."""
+
         def get_config():
             """Unnecessary now, but a deprecation warning is more trouble than it's worth."""
             return self.config
@@ -642,8 +654,8 @@ class PyFileConfigLoader(FileConfigLoader):
             __file__=self.full_filename,
         )
         conf_filename = self.full_filename
-        with open(conf_filename, 'rb') as f:
-            exec(compile(f.read(), conf_filename, 'exec'), namespace, namespace)
+        with open(conf_filename, "rb") as f:
+            exec(compile(f.read(), conf_filename, "exec"), namespace, namespace)
 
 
 class CommandLineConfigLoader(ConfigLoader):
@@ -690,13 +702,14 @@ class CommandLineConfigLoader(ConfigLoader):
         else:
             raise TypeError("Invalid flag: %r" % cfg)
 
+
 # match --Class.trait keys for argparse
 # matches:
 # --Class.trait
 # --x
 # -x
 
-class_trait_opt_pattern = re.compile(r'^\-?\-[A-Za-z][\w]*(\.[\w]+)*$')
+class_trait_opt_pattern = re.compile(r"^\-?\-[A-Za-z][\w]*(\.[\w]+)*$")
 
 _DOT_REPLACEMENT = "__DOT__"
 _DASH_REPLACEMENT = "__DASH__"
@@ -707,6 +720,7 @@ class _KVAction(argparse.Action):
 
     Always
     """
+
     def __call__(self, parser, namespace, values, option_string=None):
         if isinstance(values, str):
             values = [values]
@@ -720,11 +734,12 @@ class _KVAction(argparse.Action):
         setattr(namespace, self.dest, items)
 
 
-class _DefaultOptionDict(dict):
+class _DefaultOptionDict(dict):  # type:ignore[type-arg]
     """Like the default options dict
 
     but acts as if all --Class.trait options are predefined
     """
+
     def _add_kv_action(self, key):
         self[key] = _KVAction(
             option_strings=[key],
@@ -734,7 +749,7 @@ class _DefaultOptionDict(dict):
         )
 
     def __contains__(self, key):
-        if '=' in key:
+        if "=" in key:
             return False
         if super().__contains__(key):
             return True
@@ -759,12 +774,12 @@ class _DefaultOptionDict(dict):
 
 class _KVArgParser(argparse.ArgumentParser):
     """subclass of ArgumentParser where any --Class.trait option is implicitly defined"""
+
     def parse_known_args(self, args=None, namespace=None):
         # must be done immediately prior to parsing because if we do it in init,
         # registration of explicit actions via parser.add_option will fail during setup
         for container in (self, self._optionals):
-            container._option_string_actions = _DefaultOptionDict(
-                container._option_string_actions)
+            container._option_string_actions = _DefaultOptionDict(container._option_string_actions)
         return super().parse_known_args(args, namespace)
 
 
@@ -773,8 +788,16 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
 
     parser_class = ArgumentParser
 
-    def __init__(self, argv=None, aliases=None, flags=None, log=None, classes=(),
-                 *parser_args, **parser_kw):
+    def __init__(
+        self,
+        argv: t.Optional[t.List[str]] = None,
+        aliases: t.Optional[t.Dict[str, str]] = None,
+        flags: t.Optional[t.Dict[str, str]] = None,
+        log: t.Any = None,
+        classes: t.Optional[t.List[t.Type[t.Any]]] = None,
+        *parser_args: t.Any,
+        **parser_kw: t.Any,
+    ) -> None:
         """Create a config loader for use with argparse.
 
         Parameters
@@ -803,6 +826,7 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
         config : Config
             The resulting Config object.
         """
+        classes = classes or []
         super(CommandLineConfigLoader, self).__init__(log=log)
         self.clear()
         if argv is None:
@@ -853,13 +877,15 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
         return self.config
 
     def get_extra_args(self):
-        if hasattr(self, 'extra_args'):
+        if hasattr(self, "extra_args"):
             return self.extra_args
         else:
             return []
 
     def _create_parser(self):
-        self.parser = self.parser_class(*self.parser_args, **self.parser_kw)
+        self.parser = self.parser_class(
+            *self.parser_args, **self.parser_kw  # type:ignore[arg-type]
+        )
         self._add_arguments(self.aliases, self.flags, self.classes)
 
     def _add_arguments(self, aliases, flags, classes):
@@ -869,14 +895,14 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
         """self.parser->self.parsed_data"""
         uargs = [cast_unicode(a) for a in args]
 
-        unpacked_aliases = {}
+        unpacked_aliases: t.Dict[str, str] = {}
         if self.aliases:
             unpacked_aliases = {}
             for alias, alias_target in self.aliases.items():
                 if alias in self.flags:
                     continue
                 if not isinstance(alias, tuple):
-                    short_alias, alias = alias, None
+                    short_alias, alias = alias, None  # type:ignore[assignment]
                 else:
                     short_alias, alias = alias
                 for al in (short_alias, alias):
@@ -893,12 +919,12 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
                 if arg == k:
                     return v
                 if arg.startswith(k + "="):
-                    return v + "=" + arg[len(k) + 1:]
+                    return v + "=" + arg[len(k) + 1 :]
             return arg
 
-        if '--' in uargs:
-            idx = uargs.index('--')
-            extra_args = uargs[idx+1:]
+        if "--" in uargs:
+            idx = uargs.index("--")
+            extra_args = uargs[idx + 1 :]
             to_parse = uargs[:idx]
         else:
             extra_args = []
@@ -920,16 +946,19 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
 
 class _FlagAction(argparse.Action):
     """ArgParse action to handle a flag"""
+
     def __init__(self, *args, **kwargs):
-        self.flag = kwargs.pop('flag')
-        self.alias = kwargs.pop('alias', None)
-        kwargs['const'] = Undefined
+        self.flag = kwargs.pop("flag")
+        self.alias = kwargs.pop("alias", None)
+        kwargs["const"] = Undefined
         if not self.alias:
-            kwargs['nargs'] = 0
-        super(_FlagAction, self).__init__(*args, **kwargs)
+            kwargs["nargs"] = 0
+        super().__init__(*args, **kwargs)
 
     def __call__(self, parser, namespace, values, option_string=None):
         if self.nargs == 0 or values is Undefined:
+            if not hasattr(namespace, "_flags"):
+                namespace._flags = []
             namespace._flags.append(self.flag)
         else:
             setattr(namespace, self.alias, values)
@@ -941,15 +970,16 @@ class KVArgParseConfigLoader(ArgParseConfigLoader):
     as well as arbitrary --Class.trait value
     """
 
-    parser_class = _KVArgParser
+    parser_class = _KVArgParser  # type:ignore[assignment]
 
     def _add_arguments(self, aliases, flags, classes):
-        alias_flags = {}
+        alias_flags: t.Dict[str, t.Any] = {}
+        argparse_kwds: t.Dict[str, t.Any]
         paa = self.parser.add_argument
         self.parser.set_defaults(_flags=[])
         paa("extra_args", nargs="*")
 
-        ## An index of all container traits collected::
+        # An index of all container traits collected::
         #
         #     { <traitname>: (<trait>, <argparse-kwds>) }
         #
@@ -958,14 +988,14 @@ class KVArgParseConfigLoader(ArgParseConfigLoader):
         self.argparse_traits = argparse_traits = {}
         for cls in classes:
             for traitname, trait in cls.class_traits(config=True).items():
-                argname = '%s.%s' % (cls.__name__, traitname)
-                argparse_kwds = {'type': str}
+                argname = f"{cls.__name__}.{traitname}"
+                argparse_kwds = {"type": str}
                 if isinstance(trait, (Container, Dict)):
-                    multiplicity = trait.metadata.get('multiplicity', 'append')
-                    if multiplicity == 'append':
-                        argparse_kwds['action'] = multiplicity
+                    multiplicity = trait.metadata.get("multiplicity", "append")
+                    if multiplicity == "append":
+                        argparse_kwds["action"] = multiplicity
                     else:
-                        argparse_kwds['nargs'] = multiplicity
+                        argparse_kwds["nargs"] = multiplicity
                 argparse_traits[argname] = (trait, argparse_kwds)
 
         for keys, (value, _) in flags.items():
@@ -975,7 +1005,7 @@ class KVArgParseConfigLoader(ArgParseConfigLoader):
                 if key in aliases:
                     alias_flags[aliases[key]] = value
                     continue
-                keys = ('-' + key, '--' + key) if len(key) == 1 else ('--' + key,)
+                keys = ("-" + key, "--" + key) if len(key) == 1 else ("--" + key,)
                 paa(*keys, action=_FlagAction, flag=value)
 
         for keys, traitname in aliases.items():
@@ -984,28 +1014,28 @@ class KVArgParseConfigLoader(ArgParseConfigLoader):
 
             for key in keys:
                 argparse_kwds = {
-                    'type': str,
-                    'dest': traitname.replace(".", _DOT_REPLACEMENT),
-                    'metavar': traitname,
+                    "type": str,
+                    "dest": traitname.replace(".", _DOT_REPLACEMENT),
+                    "metavar": traitname,
                 }
                 if traitname in argparse_traits:
                     argparse_kwds.update(argparse_traits[traitname][1])
-                    if 'action' in argparse_kwds and traitname in alias_flags:
+                    if "action" in argparse_kwds and traitname in alias_flags:
                         # flag sets 'action', so can't have flag & alias with custom action
                         # on the same name
                         raise ArgumentError(
                             "The alias `%s` for the 'append' sequence "
-                            "config-trait `%s` cannot be also a flag!'"
-                            % (key, traitname))
+                            "config-trait `%s` cannot be also a flag!'" % (key, traitname)
+                        )
                 if traitname in alias_flags:
                     # alias and flag.
                     # when called with 0 args: flag
                     # when called with >= 1: alias
-                    argparse_kwds.setdefault('nargs', '?')
-                    argparse_kwds['action'] = _FlagAction
-                    argparse_kwds['flag'] = alias_flags[traitname]
-                    argparse_kwds['alias'] = traitname
-                keys = ('-' + key, '--' + key) if len(key) == 1 else ('--'+ key,)
+                    argparse_kwds.setdefault("nargs", "?")
+                    argparse_kwds["action"] = _FlagAction
+                    argparse_kwds["flag"] = alias_flags[traitname]
+                    argparse_kwds["alias"] = traitname
+                keys = ("-" + key, "--" + key) if len(key) == 1 else ("--" + key,)
                 paa(*keys, **argparse_kwds)
 
     def _convert_to_config(self):
@@ -1016,12 +1046,12 @@ class KVArgParseConfigLoader(ArgParseConfigLoader):
             if lhs == "extra_args":
                 self.extra_args = ["-" if a == _DASH_REPLACEMENT else a for a in rhs] + extra_args
                 continue
-            elif lhs == '_flags':
+            elif lhs == "_flags":
                 # _flags will be handled later
                 continue
 
             lhs = lhs.replace(_DOT_REPLACEMENT, ".")
-            if '.' not in lhs:
+            if "." not in lhs:
                 # probably a mistyped alias, but not technically illegal
                 self.log.warning("Unrecognized alias: '%s', it will have no effect.", lhs)
                 trait = None
@@ -1054,6 +1084,7 @@ class KeyValueConfigLoader(KVArgParseConfigLoader):
 
     Use KVArgParseConfigLoader
     """
+
     def __init__(self, *args, **kwargs):
         warnings.warn(
             "KeyValueConfigLoader is deprecated since Traitlets 5.0."
@@ -1081,7 +1112,7 @@ def load_pyconfig_files(config_files, path):
             next_config = loader.load_config()
         except ConfigFileNotFound:
             pass
-        except:
+        except Exception:
             raise
         else:
             config.merge(next_config)
